from .model import NanoPET
from .trainer import Trainer


__model__ = NanoPET
__trainer__ = Trainer

__authors__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]

__maintainers__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]
