from .model import SoapBpnn
from .trainer import Trainer


__model__ = SoapBpnn
__trainer__ = Trainer

__authors__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]

__maintainers__ = [
    ("Filippo Bigi <filippo.bigi@epfl.ch>", "@frostedoyster"),
]
