from .model import GAP
from .trainer import Trainer


__model__ = GAP
__trainer__ = Trainer

__authors__ = [
    ("Alexander Goscinski <alex.goscinski@posteo.de>", "@agosckinski"),
    ("Davide Tisi <davide.tisi@epfl.ch>", "@DavideTisi"),
]

__maintainers__ = [
    ("Davide Tisi <davide.tisi@epfl.ch>", "@DavideTisi"),
]
